/* 
richards.c
calculation of soil water content layer by layer taking into account soil hydrological processes 
(percolation, diffusion)  using inamic discretization level in order to taking account the change of the diffusivity and conductivity depending on soil moisture 
(after rain event the soil moisture rises suddenly causing high diffusivity and conductivity. If we assumed that these diffusivity 
 and conductivity values were valid all the day we would overestimate the velocity diffusion and percolation process) 

*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
Biome-BGCMuSo v6.2.
Copyright 2020, D. Hidy [dori.hidy@gmail.com]
Hungarian Academy of Sciences, Hungary
See the website of Biome-BGCMuSo at http://nimbus.elte.hu/bbgc/ for documentation, model executable and example input files.
*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <malloc.h>
#include "ini.h"
#include "bgc_struct.h"
#include "bgc_constants.h"
#include "bgc_func.h"
#define MIN(X, Y) (((X) < (Y)) ? (X) : (Y))
#define MAX(X, Y) (((X) > (Y)) ? (X) : (Y))   

int richards(siteconst_struct* sitec, soilprop_struct* sprop, const epconst_struct* epc, epvar_struct* epv, wstate_struct* ws, wflux_struct* wf)
{

	
	/* internal variables */
	int layer, errorCode, fr;
	double INFILT;


	/* diffusion and percolation calculation */
	double D0, D1;	        /* hydrological diffusion coefficient (m2/s) */

	double diffus, percol, wflux,  vwc_diff_max;
	double localvalue,  exponent_discretlevel;
	double soilw0, soilw1, vwc0, vwc1, vwc_diff0, vwc_diff1;
	int n_second, step, n_sec, discretlevel;

	double VWC_percol, inner, dz0, dz1;
	int dt, inORout;
	int critDepth;

	richards_struct rich;

	errorCode = 0;
	wflux=0;

	critDepth=1;   // layer thickness is 1 cm (above 1m) and 10 cm (below 1m)

	/* --------------------------------------------------------------------------------------------------------*/
	/* 1. INITALIZATION */
	
	/* time and spatial resolution */
	dt=1;	                     // 1s 


	n_second=step=0;
	n_sec=1;
	discretlevel = epc->discretlevel_Richards + 3; /* discretization level: change in VWC 1% -> timestep: second */

	/* change 10 layers to 10000 layers */
	inORout=0;
	if (richards_fineRES(inORout, critDepth, sitec, sprop, epv, wf, &rich))
	{
		if (!errorCode) 
		{
			printf("\n");
			printf("ERROR: richards_fineRES() for richards \n");
		}
		errorCode=1;
	}



	INFILT = wf->infiltPOT;
	
	
	/* --------------------------------------------------------------------------------------------------------*/	
	/* 3. CALCULATE PROCESSES WITH DYNAMIC TIME STEP (except of bottom layer) */

	while (n_second < NSEC_IN_DAY)
	{
	
		vwc_diff_max=0;
	
		/* ********************************/	
		/* 4.1 CALCULATE PROCESSES LAYER TO LAYER (except of bottom layer) */
		for (fr=0 ; fr < fineRES-1 ; fr++)
		{
			dz0=rich.dz_FR[fr];
			dz1=rich.dz_FR[fr+1];		
			/* -----------------------------*/
			/* 4.1.2. PERCOLATION */

			/* conductivity coefficient - theoretical upper limit: saturation value */
			rich.Kact_FR[fr] = rich.Ksat_FR[fr] * pow(rich.VWC_FR[fr]/rich.VWCsat_FR[fr], 2*(rich.soilB_FR[fr])+3);


			/* percolation flux - only if VWC is above field capavity*/
			if ((rich.VWC_FR[fr] - rich.VWCfc_FR[fr]) > CRIT_PREC)
			{
				inner          = rich.Ksat_FR[fr]/(dz0*(rich.VWCsat_FR[fr] - rich.VWCfc_FR[fr]*dt));
				VWC_percol     = rich.VWCfc_FR[fr] + (rich.VWC_FR[fr]  - rich.VWCfc_FR[fr]) * exp(-1*inner);
				percol = (VWC_percol - rich.VWC_FR[fr]) * water_density * dt; // kg/m2/s
			}
			else
				percol = 0;
	

			/* -----------------------------*/
			/* 4.1.3. DIFFUSION */
		
			/* diffusivity coefficient 	*/
			
	       	D0    = (((rich.soilB_FR[fr] * rich.Ksat_FR[fr] * (-100*rich.PSIsat_FR[fr]))) / 
				              rich.VWCsat_FR[fr]) * pow(rich.VWC_FR[fr]/rich.VWCsat_FR[fr] , rich.soilB_FR[fr]+2);
			
			D1    = (((rich.soilB_FR[fr+1] * rich.Ksat_FR[fr+1] * (-100*rich.PSIsat_FR[fr+1]))) / 
				               rich.VWCsat_FR[fr+1]) * pow(rich.VWC_FR[fr+1]/rich.VWCsat_FR[fr+1] , rich.soilB_FR[fr+1]+2);

			rich.Dact_FR[fr]  = (D0 + D1) /2;
	
		
			/* diffusion flux - only if target layer is NOT saturated */
			diffus = (rich.VWC_FR[fr] - rich.VWC_FR[fr+1])/((dz0+dz1)/2) 	* rich.Dact_FR[fr]  * water_density;


		
			/* -----------------------------*/
			/* 4.1.5. FINDING THE MAXIMAL CHANGE to determine discretization step */
		
		
			soilw0 = rich.soilw_FR[fr]   - (percol + diffus);
			vwc0   = soilw0 / dz0 / water_density;
			vwc_diff0 = fabs(rich.VWC_FR[fr]  - vwc0);

			/* bottom layer is special: i+1 layer is he boundary layer of which water content does not change */
			if (fr < fineRES-1)
			{
				soilw1 = rich.soilw_FR[fr+1] + (percol + diffus);
				vwc1   = soilw1 / dz0 / water_density;
				vwc_diff1 = fabs(rich.VWC_FR[fr+1] - vwc1);
		
				if (vwc_diff_max < vwc_diff0 || vwc_diff_max < vwc_diff1) vwc_diff_max = MAX(vwc_diff0, vwc_diff1);			
			}
			else
			{
				if (vwc_diff_max < vwc_diff0) vwc_diff_max = vwc_diff0;
			}
		
			/* -----------------------------*/
			/* 4.1.6. UPDATING STATE VARIABLES    */

	
			rich.soilw_FR[fr]   -= wflux;
			rich.VWC_FR[fr]      = rich.soilw_FR[fr]   / dz0 / water_density;

			rich.soilw_FR[fr+1] += wflux;
			rich.VWC_FR[fr+1]    = rich.soilw_FR[fr+1] / dz1 / water_density;
			
			rich.percol_FR[fr]   += percol * n_sec;
			rich.diffus_FR[fr]   += diffus * n_sec;

		}
		n_second += n_sec;

		/* ********************************/	
		/* 4.2. CALCULATION OF DISCRETE LEVEL: if the maximal change of VWC is not equal to 0 (greater than) -> discretlevel is the function of local value of the change */
	
		if (vwc_diff_max > CRIT_PREC)
		{
			localvalue=floor(log10(vwc_diff_max));
			if (localvalue + discretlevel < 0)
				exponent_discretlevel = fabs(localvalue + discretlevel);
			else
				exponent_discretlevel = 0;

			n_sec = (int) pow(10, exponent_discretlevel);
		}
		else
		{
			n_sec = NSEC_IN_DAY - n_second;
		}
		vwc_diff_max = 0;

		/* ********************************/
		/* 4.3 CALCULATION OF CONDUCTANCE AND DIFFUSION PARAMETERS */
		
		/* 4.3.1 hydrCONDUCT and hydrDIFFUS at the beginning of the calculation  */
		if (step == 0)
		{	for (layer = 0; layer < N_SOILLAYERS; layer++)
			{
				//epv->hydr_conductSTART[layer] = hydrCONDUCT[layer];
				//epv->hydr_diffusSTART[layer] = hydrDIFFUS[layer];
			}
		}
		step=step+1;	
	}

	/* change 140 layers to 10 layers */
	inORout=1;
	if (richards_fineRES(inORout, critDepth, sitec, sprop, epv, wf, &rich))
	{
		if (!errorCode) 
		{
			printf("\n");
			printf("ERROR: richards_fineRES() for richards \n");
		}
		errorCode=1;
	}	

	return (errorCode);
}


int richards_fineRES(int inORout, double critDepth, siteconst_struct* sitec, soilprop_struct* sprop, epvar_struct* epv, wflux_struct* wf, richards_struct* rich)
{	
	int nlayer, fr, errorCode;
	double depth, dz;
	int resol1, resol2;
	nlayer=errorCode = 0;

	
	resol1=1; 
	resol2=10; 

	/* new VWC values after Richards-method */
	if (inORout == 1)
	{
		for (nlayer = 0; nlayer < N_SOILLAYERS-1; nlayer++) epv->VWC[nlayer] = 0;
	}

	depth=0.01;  // depth of the first layer in fine resolution
	

	for (fr=0 ; fr < fineRES ; fr++)
	{
		if (depth - critDepth > CRIT_PREC)
			dz=0.1;
		else
			dz=0.01;

		if (depth > sitec->soillayer_depth[0])
		{
			if (depth > sitec->soillayer_depth[1])
			{	
				if (depth > sitec->soillayer_depth[2])
				{
					if (depth > sitec->soillayer_depth[3])
					{
						if (depth > sitec->soillayer_depth[4])
						{
							if (depth > sitec->soillayer_depth[5])
							{
								if (depth > sitec->soillayer_depth[6])
								{
									if (depth > sitec->soillayer_depth[7])
									{
										if (depth > sitec->soillayer_depth[8])
										{
												nlayer = 9;
										}
										else 
										{
											nlayer = 8;
										}
									}
									else 
									{
										nlayer = 7;
									}
								}
								else 
								{
									nlayer = 6;
								}
							}
							else 
							{
								nlayer = 5;
							}
						}
						else 
						{
							nlayer = 4;
						}
					}
					else
					{
						nlayer = 3;
					}
				}
				else 
				{
					nlayer = 2;
				}	
			}
			else 
			{
				nlayer = 1;
			}
		}
		else
		{
			nlayer = 0;
		}
		
		/* input for Richards */
		if (inORout == 0)
		{
			rich->VWC_FR[fr]      = epv->VWC[nlayer];
			rich->VWCfc_FR[fr]    = sprop->VWCfc[nlayer];
			rich->VWCsat_FR[fr]   = sprop->VWCsat[nlayer];
			rich->PSIsat_FR[fr]   = sprop->PSIsat[nlayer];
			rich->soilB_FR[fr]    = sprop->soilB[nlayer];
			rich->Ksat_FR[fr]     = sprop->hydrCONDUCTsat[nlayer];
			rich->Dsat_FR[fr]     = sprop->hydrDIFFUSsat[nlayer];
			rich->percol_FR[fr]   = 0;
			rich->diffus_FR[fr]   = 0;
			rich->soilw_FR[fr]    = rich->VWC_FR[fr]*dz*water_density;
			rich->dz_FR[fr]       = dz;
		}

		/* output from Richards */
		else
		{
			/* averaged VWC */
			epv->VWC[nlayer]      += rich->VWC_FR[fr] * (dz/sitec->soillayer_thickness[nlayer]);

			/* percolation and diffusion from boundary fineRES layer */
			if (sitec->soillayer_depth[nlayer]-dz-depth < CRIT_PREC)
			{
				wf->soilw_percolated[nlayer]   = rich->percol_FR[fr];
				wf->soilw_diffused[nlayer] = rich->diffus_FR[fr];
			}
		}

	
			
		depth+=dz;  // layer thickness below critical depth: 1	 cm
		
	}

	


	return (errorCode);
}